package org.homebrew;

import java.io.*;
import java.util.*;
import java.awt.*;
import java.awt.image.*;
import java.net.*;

import javax.media.*;
import javax.tv.xlet.*;

import org.bluray.ui.event.HRcEvent;
import org.dvb.event.UserEvent;
import org.dvb.event.EventManager;
import org.dvb.event.UserEventListener;
import org.dvb.event.UserEventRepository;
import org.havi.ui.*;
import org.dvb.ui.*;


public class MyXlet extends Container implements Xlet, Runnable,UserEventListener {

    static final float PI = 3.141593F;
    static final int RENDERWIDTH = 800;
    static final int RENDERHEIGHT = 550;
	static final int MAPWIDTH = 512;
	static final int MAPWIDTH_MIN_1 = MAPWIDTH - 1;
	static final int MAPSIZE = MAPWIDTH * MAPWIDTH;
	static final int MAPSHIFT = 9;
	
	private HScene scene;
	private XletContext context;
	private final ArrayList messages = new ArrayList();
	
	private Thread myWorkerThread;
	private boolean running = true;
	
	private static final long serialVersionUID = 4761178503523947426L;
	private Font font;
	
	private DVBBufferedImage buffer = null;
	private Graphics2D bufferG;
	private int sinTable[];
    private int cosTable[];
    private int heightMap[];
    private int texelMap[];
    private int renderBuffer[];
    private int colorLut[];
    
    private int playerX = 0;
    private int playerY = 0;
    private int playerZ = 350 << 12;
    private int playerDistance = 277 << 12;
    private int playerSpeed = 10 << 12;
    private int playerAngle = 0;
    private int angleSpeed = 0;
    private int viewportHeight = 100 << 12;

    private boolean initialized = false;
	private String error = null;
	private String line = "0";
	private int frame=0;
		
	public void initXlet(XletContext context) {
		this.context = context;

		scene = HSceneFactory.getInstance().getDefaultHScene();

		try
		{

			setSize(1920, 1080); // BD screen size
			scene.add(this, BorderLayout.CENTER);

			//font
			font = new Font(null, Font.PLAIN, 36);
						
			//add input  event			
			UserEventRepository userEventRepo = new UserEventRepository("x");
	        userEventRepo.addAllArrowKeys();
	        userEventRepo.addAllColourKeys();
	        userEventRepo.addAllNumericKeys();
	        userEventRepo.addKey(HRcEvent.VK_ENTER);
	        userEventRepo.addKey(HRcEvent.VK_POPUP_MENU);

			EventManager.getInstance().addUserEventListener(this, userEventRepo);
			
		} catch (Exception e) {
			error = e.getMessage();
		}
		scene.validate();
	}

	public void startXlet() {
		setVisible(true);
		scene.setVisible(true);
		requestFocus();
		
		myWorkerThread = new Thread(this);
		myWorkerThread.start();
	}

	public void pauseXlet() {
		setVisible(false);
	}

	public void destroyXlet(boolean unconditional) {
		scene.remove(this);
		scene = null;
	}

	public void update(Graphics g)
    {
    }
    
    private void initTables()
    {
        sinTable = new int[3600];
        cosTable = new int[3600];
        for (int i=0; i<3600; i++) {
            sinTable[i] = MathFP.toFP((float)Math.sin(2f*3.141593f*i/3600f));
            cosTable[i] = MathFP.toFP((float)Math.cos(2f*3.141593f*i/3600f));
        }
        colorLut = new int[65536];
        for (int i=0; i<256; i++) {
        	for (int j=0; j<256; j++) {
        		int idx = (i << 8) + j;
        		float c = (float)i * (float)j/256f;
        		colorLut[idx] = (int)Math.round(c);
        	}
        }
    }
    
	private void initImages() {
		try
		{
			ImageLoader texLoader = new ImageLoader("/land.jpg", true, this);
            texLoader.load();
            while(!texLoader.ready()) 
                try
                {
                    Thread.sleep(100L);
                }
                catch(Exception _ex) { }
            texelMap = (int[])texLoader.getPixels();
           
            ImageLoader heightLoader = new ImageLoader("/height.jpg", false, this);
            heightLoader.load();
            while(!heightLoader.ready()) 
                try
                {
                    Thread.sleep(100L);
                }
                catch(Exception _ex) { }
            heightMap = new int[MAPSIZE];
            int abyte0[] = (int[])heightLoader.getPixels();
            int k = 0;
            do {
            	int aByte = (0x000000FF & ((int)abyte0[k]));
                heightMap[k] = aByte;
            } while(++k < MAPSIZE);
            
            renderBuffer = new int[RENDERWIDTH * RENDERHEIGHT];
        			
            
		} catch (Exception e) {
			error = e.getMessage();
		}
	}
	
	public void render2(int eyeX, int eyeY, int eyeZ, int arc, int dpk)
    {
    	int k=0;
        for(int i=0; i<RENDERHEIGHT; i++) {
			for (int j=0; j<RENDERWIDTH; j++) {
				renderBuffer[k++] = 0xFF0000FF;
			}
        }

        int colAngle = arc - 400; // view from angle-30 to angle+30
        if(colAngle < 0)
            colAngle += 3600;
        int col = 0;
        int pitch = eyeZ - viewportHeight;
        int invdpk = MathFP.div(1 << 12, dpk);
        do 			  // COL iteration
        {
            int j3 = RENDERHEIGHT * RENDERWIDTH - RENDERWIDTH + col;
            int cosA = cosTable[colAngle];
            int sinA = sinTable[colAngle];
            int z = eyeZ;
            int x = eyeX;
            int y = eyeY;
            int dz = 0;
            int m = MathFP.div(pitch, dpk); 
            //int j = 0; // row
            for (int i=0; i<800-256; i++) 	// abscis iteration
            {
                y += sinA;
                x += cosA;
                z -= m;
                dz += invdpk;
                int xm = (x >> 12) & MAPWIDTH_MIN_1;  // mod 512
                int ym = (y >> 12) & MAPWIDTH_MIN_1;  // mod 512
                int idx = (ym << MAPSHIFT) + xm;
                int hi = heightMap[idx] << 12;
                while (hi > z && j3 > 0) {
                	int texel = texelMap[idx];
                	renderBuffer[j3] = texel;
                	j3 -= RENDERWIDTH;
                	m -= invdpk;
                	z += dz;
                	//j++; // move up a screen row
                } 
                if(j3 < 0)
                    break;
            }
            for (int i=0; i<256; i++) 	// abscis iteration
            {
                y += sinA;
                x += cosA;
                z -= m;
                dz += invdpk;
                int xm = (x >> 12) & MAPWIDTH_MIN_1;  // mod 512
                int ym = (y >> 12) & MAPWIDTH_MIN_1;  // mod 512
                int idx = (ym << MAPSHIFT) + xm;
                int hi = heightMap[idx] << 12;
                while (hi > z && j3 > 0) {
                	int texel = blend(texelMap[idx],0xFF0000FF,i);
                	renderBuffer[j3] = texel;
                	j3 -= RENDERWIDTH;
                	m -= invdpk;
                	z += dz;
                	//j++; // move up a screen row
                } 
                if(j3 < 0)
                    break;
            }
            if(++colAngle >= 3600)
                colAngle = 0;
        } while(++col < RENDERWIDTH);
    }
	
	public void run()
	{
		initTables();
		initImages();
		Runtime runtime = Runtime.getRuntime();
		initialized = true;
		while(running)
		{
			 try
	                {
        	            myWorkerThread.sleep(20);
                	}
	                catch(Exception _ex) { }
	        playerAngle += angleSpeed;
	        if (playerAngle < 0)
	        	playerAngle += 3600;
	        else if (playerAngle >= 3600)
	        	playerAngle -= 3600;
	        int cos = cosTable[playerAngle];
            int sin = sinTable[playerAngle];
            playerX += MathFP.mul(cos, playerSpeed);
            playerY += MathFP.mul(sin ,playerSpeed);
	
			render2(playerX, playerY, playerZ, playerAngle, playerDistance);
			//runtime.gc();
			
			frame++;
			
			repaint();
		}
	}
	
	
	public void userEventReceived(UserEvent evt)
	{
		if (evt.getType() == HRcEvent.KEY_PRESSED)
		{
			switch(evt.getCode())
			{
				case HRcEvent.VK_POPUP_MENU:
					playerZ += 10 << 12;
	                break;
	                
				case HRcEvent.VK_ENTER:
					playerZ -= 10 << 12;
	                break;

	            case HRcEvent.VK_LEFT:
	            	//if ((playerAngle -= 50) < 0)
                	//	playerAngle += 3600;
                	angleSpeed -= 5;
	                break;

	            case HRcEvent.VK_RIGHT:
	            	//if ((playerAngle += 50) >= 3600)
                	//	playerAngle -= 3600;
                	angleSpeed += 5;
	                break;

	            case HRcEvent.VK_UP:
	            	playerSpeed += 1 << 12;
	                break;

	            case HRcEvent.VK_DOWN:
	            	playerSpeed -= 1 << 12;
	                break;
	                
	            
	            case HRcEvent.VK_0:
	            case HRcEvent.VK_1:
	            case HRcEvent.VK_2:
	            case HRcEvent.VK_3:
	            case HRcEvent.VK_4:
	            case HRcEvent.VK_5:
	            case HRcEvent.VK_6:
	            case HRcEvent.VK_7:
	            case HRcEvent.VK_8:
	            case HRcEvent.VK_9:
	            	messages.add("HRcEvent.VK_numer : " + evt.getCode());
	            	break;
	            		            
			}
		}		
	}
    
    public void paint(Graphics g)
	{
		checkOffscreenImage();

		Graphics offG = bufferG;
		
		offG.setColor(new Color(255, 255, 255));
		offG.fillRect(0, 0, 400, 140);
		
		//draw
		offG.setFont(font);
				
		//draw text
		offG.setColor(new Color(0, 0, 0));
        int top = 100;
		
		if (error == null)
        	offG.drawString("frame="+frame, 120, 100);
        else
        	offG.drawString("error="+error+" line="+line,100,100);
        
        if (initialized)
        	buffer.setRGB(550,154,RENDERWIDTH,RENDERHEIGHT,renderBuffer,0,RENDERWIDTH);
        
         g.drawImage(buffer, 0, 0, null);
    
    }
	
	 private void checkOffscreenImage() 
	 {
		    Dimension d = getSize();
		    if (buffer == null || buffer.getWidth(null) != d.width
		        || buffer.getHeight(null) != d.height)
		    {
		    		buffer = new DVBBufferedImage(d.width, d.height);
					org.dvb.ui.DVBGraphics dvbgraphics = buffer.createGraphics();
        			bufferG = (Graphics2D)dvbgraphics;
        			bufferG.setComposite(AlphaComposite.Src);
        			
		    }
	 }
	 
	 private final int blend(int c1, int c2, int f) {
	 	int r1 = (c1 & 0x00FF0000) >> 16;
	 	int r2 = (c2 & 0x00FF0000) >> 16;
	 	int r = (colorLut[(r1 << 8) + (255 - f)] + colorLut[(r2 << 8) + f]) << 16;
	 	
	 	int g1 = (c1 & 0x0000FF00) >> 8;
	 	int g2 = (c2 & 0x0000FF00) >> 8;
	 	int g = (colorLut[(g1 << 8) + (255 - f)] + colorLut[(g2 << 8) + f]) << 8;
	 	
	 	int b1 = (c1 & 0x000000FF);
	 	int b2 = (c2 & 0x000000FF);
	 	int b = colorLut[(b1 << 8) + (255 - f)] + colorLut[(b2 << 8) + f];
	 	
	 	return 0xFF000000 + r + g + b;
	 	
	 }
}